# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Sample
                                 A QGIS plugin
 QGIS Sample Plugin
                              -------------------
        begin                : 2020-06-01
        git sha              : $Format:%H$
        copyright            : (C) 2020 by MIERUNE Inc.
        email                : info@mierune.co.jp
        license              : GNU General Public License v2.0
 ***************************************************************************/
"""

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from qgis.core import *
from qgis.gui import *

from .Png2TiffBase import Png2TiffBase
from .PngTranslator import PngTranslator
from .. import settings


class Png2Tiff:
    def __init__(self, iface):
        self.iface = iface
        self.canvas = self.iface.mapCanvas()

        self.dlg = Png2TiffBase()
        self.init_gui()

    # キャンセルクリック
    def dlg_cancel(self):
        self.dlg.close()

    def init_gui(self):
        # 読み込み対象レイヤーはラスターのみ
        self.dlg.inputLayerComboBox.setFilters(
            QgsMapLayerProxyModel.RasterLayer)

        # UIの状態変化を検知して表示を更新
        self.dlg.inputLayerComboBox.currentIndexChanged.connect(
            self.refresh_gui_by_user_input)
        self.dlg.outputProjectionSelectionWidget.crsChanged.connect(
            self.refresh_gui_by_user_input)
        self.refresh_gui_by_user_input()

        # 投影法プルダウンのデフォルトはプロジェクトCRSに
        project_crs = QgsProject.instance().crs()
        self.dlg.outputProjectionSelectionWidget.setCrs(project_crs)

        # 1ピクセルあたりの地理的距離スピンボックスの小数点以下桁数をセット
        self.dlg.inputLayerdistXSpinBox.setDecimals(
            settings.PX_DIST_DECIMAL_DIGITS)
        self.dlg.inputLayerdistYSpinBox.setDecimals(
            settings.PX_DIST_DECIMAL_DIGITS)

        # 基準標高・標高係数の小数点以下桁数をセット
        self.dlg.outputLowestElevationSpinBox.setDecimals(
            settings.HEIGHT_DECIMAL_DIGITS)
        self.dlg.outputElevationMultiplySpinBox.setDecimals(
            settings.HEIGHT_MULTIPLIER_DECIMAL_DIGITS)

        # ボタンに処理を紐付け
        self.dlg.pushButton_run.clicked.connect(self.on_run_button_pushed)
        self.dlg.pushButton_cancel.clicked.connect(self.dlg_cancel)

        # 保存ファイル形式はTIFF
        self.dlg.outputFileWidget.setFilter("TIFF画像形式(*.tif)")

    def refresh_gui_by_user_input(self):
        """[summary]
        ユーザーの入力に応じてGUIの表示を更新する
        """

        rlayer = self.dlg.inputLayerComboBox.currentLayer()
        if rlayer is None:
            self.dlg.inputLayerWidthLabel.setText("幅:0px")
            self.dlg.inputLayerHeightLabel.setText("高さ:0px")
            self.dlg.inputLayerOriginXSpinBox.setValue(0.0)
            self.dlg.inputLayerOriginYSpinBox.setValue(0.0)
            self.dlg.inputLayerdistXSpinBox.setValue(0.1)
            self.dlg.inputLayerdistYSpinBox.setValue(0.1)
            self.dlg.outputLowestElevationSpinBox.setValue(0.0)
            self.dlg.outputElevationMultiplySpinBox.setValue(0.001)
            self.dlg.inputInfocsvExistLabel.setText("")
        else:
            geo_info = {
                "height": rlayer.height(),
                "width": rlayer.width(),
                "x_origin": rlayer.dataProvider().extent().xMinimum(),
                "y_origin": rlayer.dataProvider().extent().yMaximum(),
                "x_dist": rlayer.rasterUnitsPerPixelX(),
                "y_dist": rlayer.rasterUnitsPerPixelY()
            }
            self.dlg.inputLayerWidthLabel.setText(f"幅:{geo_info['width']}px")
            self.dlg.inputLayerHeightLabel.setText(
                f"高さ:{geo_info['height']}px")
            self.dlg.inputLayerOriginXSpinBox.setValue(geo_info["x_origin"])
            self.dlg.inputLayerOriginYSpinBox.setValue(geo_info["y_origin"])
            self.dlg.inputLayerdistXSpinBox.setValue(geo_info["x_dist"])
            self.dlg.inputLayerdistYSpinBox.setValue(geo_info["y_dist"])

            # 基準標高・最低標高
            elev_lowest, elev_multiplier = self.get_elev_params_from_csv(
                rlayer.source())
            self.dlg.outputLowestElevationSpinBox.setValue(float(elev_lowest))
            self.dlg.outputElevationMultiplySpinBox.setValue(
                float(elev_multiplier))
            self.dlg.inputInfocsvExistLabel.setText(
                self.infocsv_exist_label_text(self.get_infocsv_path(rlayer.source())))

        # エラーがあればメッセージを表示し実行ボタンを無効化
        error_msg = self.has_error_in_user_input()
        self.dlg.errorMessageLabel.setText(error_msg)
        self.dlg.pushButton_run.setEnabled(error_msg == "")

    def infocsv_exist_label_text(self, csvfile_path: str) -> str:
        if csvfile_path is None:
            return "※info.csvが入力画像と同じフォルダに見つかりませんでした"
        else:
            return "※info.csvからパラメータが自動入力されました"

    def get_infocsv_path(self, png_path: str) -> str:
        """
        本プラグイン準拠のフォルダ構造が維持されている場合
        info.csvへの絶対パスを返す
        なければNoneを返す

        Args:
            png_path (str): [description]

        Returns:
            str: csvファイルへのパス
        """
        import os
        dir_of_png = os.path.dirname(png_path)
        # このプラグインで出力したままならそこにinfo.csvがある
        csvfile_path = os.path.join(dir_of_png, "info.csv")
        if os.path.exists(csvfile_path):
            return csvfile_path
        else:
            return None

    def get_elev_params_from_csv(self, png_path: str) -> tuple:
        """
        本プラグイン準拠のフォルダ構造が維持されている場合
        info.csvを読み込んで基準標高・標高係数を読み取って返す
        なければ(0, 0)を返す

        Args:
            png_path (str): PNGファイルへの絶対パス

        Returns:
            tuple (float、float): 基準標高・標高係数
        """
        import os
        import csv

        # このプラグインで出力したままならそこにinfo.csvがある
        csvfile_path = self.get_infocsv_path(png_path)
        if csvfile_path:
            with open(csvfile_path) as f:
                reader = csv.DictReader(f)
                datas = [row for row in reader]
                if len(datas) > 0:
                    elev_lowest = datas[0].get("elev_lowest")
                    elev_multiplier = datas[0].get("elev_multiplier")
                    return elev_lowest, elev_multiplier
        return 0, 0

    def has_error_in_user_input(self) -> str:
        """[summary]
        ユーザーのUIへの入力にエラーがあるかどうか判定する
        エラーがあればメッセージ文字列を、なければ空文字列を返す

        Returns:
            str: エラーメッセージ
        """

        rlayer = self.dlg.inputLayerComboBox.currentLayer()
        if rlayer is None:
            return "プロジェクトに有効なレイヤーが存在しません"
        if not rlayer.bandCount() == 1 or \
                not rlayer.dataProvider().dataType(1) == Qgis.DataType.UInt16:
            return "入力画像は16bitグレースケールでなければなりません"

        selected_epsg_code = self.dlg.outputProjectionSelectionWidget.crs().authid()
        if selected_epsg_code == "":
            return "座標系を指定する必要があります"
        if not int(selected_epsg_code.split(":")[-1]) in settings.SUPPORTED_EPSG_CODES:
            return "座標系は平面直角座標系でなければなりません"
        return ""

    def on_run_button_pushed(self):
        if self.dlg.outputFileWidget.filePath() == "":
            QMessageBox.information(None, "エラー", "出力先を指定してください")
            return

        rlayer = self.dlg.inputLayerComboBox.currentLayer()
        png_translator = PngTranslator(
            rlayer.source(),
            x_origin=self.dlg.inputLayerOriginXSpinBox.value(),
            y_origin=self.dlg.inputLayerOriginYSpinBox.value(),
            x_dist=self.dlg.inputLayerdistXSpinBox.value(),
            y_dist=self.dlg.inputLayerdistYSpinBox.value()
        )

        # translate用のパラメータ
        src_min, src_max = 0, 65535
        elevation_lowest = self.dlg.outputLowestElevationSpinBox.value()
        elevation_multiplier = self.dlg.outputElevationMultiplySpinBox.value()
        dst_min = round(elevation_lowest, settings.HEIGHT_DECIMAL_DIGITS)
        dst_max = round(elevation_lowest + elevation_multiplier * (src_max - src_min),
                        settings.HEIGHT_MULTIPLIER_DECIMAL_DIGITS)
        png_translator.translate_to_tiff(
            output_path=self.dlg.outputFileWidget.filePath(),
            scaleParams=[[src_min, src_max, dst_min, dst_max]],
            outputSRS=self.dlg.outputProjectionSelectionWidget.crs().authid()
        )

        QMessageBox.information(None, "処理完了", "ファイルの書き出しが完了しました")
