# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Sample
                                 A QGIS plugin
 QGIS Sample Plugin
                              -------------------
        begin                : 2018-12-23
        git sha              : $Format:%H$
        copyright            : (C) 2018 by MIERUNE Inc.
        email                : info@mierune.co.jp
        license              : GNU General Public License v2.0
 ***************************************************************************/
"""

import gc

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from qgis.core import *
from qgis.gui import *

from .. import settings


class RubberbandMapTool(QgsMapTool):
    def __init__(self, iface, callback, tilesize, centerMode):
        """[summary]
        矩形選択マップツール
        Args:
            iface : QgisInterface
            callback (function): 矩形確定時に実行する関数
            tilesize ([float, float]): PNGタイル1枚の地理的距離xy

        Returns:
            [type]: [description]
        """
        QgsMapTool.__init__(self, iface.mapCanvas())
        self.iface = iface
        self.callback = callback
        self.canvas = iface.mapCanvas()
        self.tilesize = tilesize
        self.centerMode = centerMode

        # 直前のマップツールを記憶しておいて矩形定義後に戻す
        self.previous_map_tool = iface.mapCanvas().mapTool()

        # ドラッグ開始時のマップキャンバス座標(QPoint)を保存する変数
        self.startpoint = None

        # ドラッグ中に描画されるグリッドを保存する変数
        self.grids = []

        # 描画されるラバーバンド
        self.rubberband = QgsRubberBand(
            self.canvas, QgsWkbTypes.PolygonGeometry)

        return None

    # クリック時の動作
    def canvasPressEvent(self, e):
        self.startpoint = QPoint(e.pos().x(), e.pos().y())
        return None

    # ドラッグ時の動作
    def canvasMoveEvent(self, e):
        # ドラッグ中はラバーバンドを常に再描画する
        self.rubberband.reset(QgsWkbTypes.PolygonGeometry)

        currentpoint = QPoint(e.pos().x(), e.pos().y())
        vertexes = self.get_current_vertexes(currentpoint)
        for v in vertexes:
            self.rubberband.addPoint(v, True)

        self.rubberband.show()

        if self.startpoint:
            self.show_tile_grids(currentpoint)
        return None

    # クリックボタンが離された時の動作
    def canvasReleaseEvent(self, e):
        # クリックボタン離されたときのカーソルの座標を保存
        endpoint = QPoint(e.pos().x(), e.pos().y())

        # クリック開始時・離した時の座標を現在のCRS座標に変換
        if self.centerMode:
            local_startpoint = self.startpoint
            # if type(self.startpoint.x) == int:
            #    local_startpoint = QPoint(self.startpoint.x, self.startpoint.y)
            point_ext = endpoint - local_startpoint
            point_sta = local_startpoint - point_ext
            point_end = local_startpoint + point_ext
            start_coordinates = self.canvas.getCoordinateTransform().toMapCoordinates(point_sta.x(), point_sta.y())
            end_coordinates = self.canvas.getCoordinateTransform().toMapCoordinates(point_end.x(), point_end.y())
            lefttop_xy, rightbottom_xy = self.sort_as_lefttop_rightbottom(
                 start_coordinates, end_coordinates)
        else:
            start_coordinates = self.canvas.getCoordinateTransform(
            ).toMapCoordinates(self.startpoint.x(), self.startpoint.y())
            end_coordinates = self.canvas.getCoordinateTransform(
            ).toMapCoordinates(endpoint.x(), endpoint.y())
            lefttop_xy, rightbottom_xy = self.sort_as_lefttop_rightbottom(
                start_coordinates, end_coordinates)

        # ドラッグ終了点が属する象限計算
        quadrant = self.calc_quadrant(start_coordinates, end_coordinates)
        # 生成予定のタイル数計算
        x_grids_count, y_grids_count, x_sign, y_sign = self.calc_xy_grids_count(endpoint)
        grids_count = abs(x_grids_count) * abs(y_grids_count)

        if self.centerMode:
            center_coordinates = self.canvas.getCoordinateTransform().toMapCoordinates(self.startpoint.x(), self.startpoint.y())
            ext_coordinates = QgsPoint(self.tilesize[0] * x_grids_count * 0.5, self.tilesize[1] * y_grids_count * 0.5)
            start_coordinates = QgsPoint(center_coordinates.x() - ext_coordinates.x(), center_coordinates.y() - ext_coordinates.y())
            end_coordinates = QgsPoint(center_coordinates.x() + ext_coordinates.x(), center_coordinates.y() + ext_coordinates.y())
            lefttop_xy, rightbottom_xy = self.sort_as_lefttop_rightbottom(
                 start_coordinates, end_coordinates)

        # コールバック関数を実行
        self.callback(lefttop_xy, rightbottom_xy, quadrant, grids_count)

        # 処理が終了したのでマップツールを戻して各種変数を初期化
        self.canvas.setMapTool(self.previous_map_tool)
        self.startpoint = None
        self.rubberband.reset(QgsWkbTypes.PolygonGeometry)
        self.reset_all_grids()
        return None

    # ラバーバンドで定義された矩形の4頂点のCRS座標を返す関数
    def get_current_vertexes(self, currentpoints):
        if self.startpoint and currentpoints:
            if self.centerMode:
                local_startpoint = self.startpoint
                # if type(self.startpoint.x) == int:
                #    local_startpoint = QPoint(self.startpoint.x, self.startpoint.y)
                point_extent = currentpoints - local_startpoint
                point_sta = local_startpoint - point_extent
                point_end = local_startpoint + point_extent
                p1 = self.canvas.getCoordinateTransform().toMapCoordinates(point_sta.x(), point_sta.y())
                p2 = self.canvas.getCoordinateTransform().toMapCoordinates(point_end.x(), point_sta.y())
                p3 = self.canvas.getCoordinateTransform().toMapCoordinates(point_end.x(), point_end.y())
                p4 = self.canvas.getCoordinateTransform().toMapCoordinates(point_sta.x(), point_end.y())
                return [p1, p2, p3, p4]
            else:
                p1 = self.canvas.getCoordinateTransform().toMapCoordinates(
                    self.startpoint.x(), self.startpoint.y())
                p2 = self.canvas.getCoordinateTransform().toMapCoordinates(
                    currentpoints.x(), self.startpoint.y())
                p3 = self.canvas.getCoordinateTransform().toMapCoordinates(
                    currentpoints.x(), currentpoints.y())
                p4 = self.canvas.getCoordinateTransform().toMapCoordinates(
                    self.startpoint.x(), currentpoints.y())
                return [p1, p2, p3, p4]
        return []

    def reset_all_grids(self):
        """保持しているグリッドのインスタンスすべてのジオメトリを消去する(キャンバスから消去する)
        """
        for grid in self.grids:
            grid.reset(QgsWkbTypes.PolygonGeometry)

    def show_tile_grids(self, currentpoint: QPoint):
        """[summary]
        ドラッグ時にマウスが動くたびに実行される関数
        グリッドを全消去してから再描画する
        Args:
            currentpoint (QPoint): ドラッグ中のキャンバス上の座標
        """

        # 描画前に保持しているグリッドすべてのジオメトリをリセットする（キャンバスから消去）
        self.reset_all_grids()

        local_startpoint = self.startpoint
        local_currentpoint = currentpoint
        if self.centerMode:
            # if type(self.startpoint.x) == int:
            #    local_startpoint = QPoint(self.startpoint.x, self.startpoint.y)
            point_extent = currentpoint - local_startpoint
            local_currentpoint = local_startpoint + point_extent
            local_startpoint = local_startpoint - point_extent

        # 描画すべきグリッドの数を計算、その分処理を繰り返す
        x_count, y_count, x_sign, y_sign = self.calc_xy_grids_count(local_currentpoint)

        # グリッド数が多すぎる場合、パフォーマンスの低下を避けるため描画を停止
        if abs(x_count) * abs(y_count) > settings.MAX_SHOWN_GRIDS_COUNT:
            self.iface.messageBar().pushInfo(
                'お知らせ', f'グリッド数が{settings.MAX_SHOWN_GRIDS_COUNT}を超えたため描画を停止します')
            return

        # グリッド生成処理のループのためのカウンター
        grids_counter = 0

        if self.centerMode:
            center_coordinates = self.canvas.getCoordinateTransform().toMapCoordinates(self.startpoint.x(), self.startpoint.y())
            start_coordinates = QgsPointXY(
                center_coordinates.x() - self.tilesize[0] * x_count * 0.5, 
                center_coordinates.y() - self.tilesize[1] * y_count * 0.5)
            local_startpoint = self.toCanvasCoordinates(start_coordinates)

        for y in range(abs(y_count)):
            for x in range(abs(x_count)):
                # 処理開始時にカウンター増加
                grids_counter += 1

                # クリックの開始地点の地理座標
                start_coordinates = self.canvas.getCoordinateTransform(
                ).toMapCoordinates(local_startpoint.x(), local_startpoint.y())
                start_coordinates = QgsPointXY(start_coordinates.x(), start_coordinates.y())

                # 各グリッドの座標を計算
                # 開始地点 + グリッドのxy番地 * グリッド(タイル)1枚あたりの地理的サイズ
                grid_vertexes = [
                    QgsPointXY(start_coordinates.x() + self.tilesize[0] * x * x_sign,
                               start_coordinates.y() + self.tilesize[1] * y * y_sign),
                    QgsPointXY(start_coordinates.x() + self.tilesize[0] * (x + 1) * x_sign,
                               start_coordinates.y() + self.tilesize[1] * y * y_sign),
                    QgsPointXY(start_coordinates.x() + self.tilesize[0] * (x + 1) * x_sign,
                               start_coordinates.y() + self.tilesize[1] * (y + 1) * y_sign),
                    QgsPointXY(start_coordinates.x() + self.tilesize[0] * x * x_sign,
                               start_coordinates.y() + self.tilesize[1] * (y + 1) * y_sign)
                ]

                # 一度生成したグリッドは一定数インスタンスを保持し続ける
                if grids_counter > len(self.grids):
                    # 保持している数以上のグリッドを表示する場合新たに生成する
                    grid_rband = QgsRubberBand(
                        self.canvas, QgsWkbTypes.PolygonGeometry)
                    # 生成したグリッドはすべて配列に保存
                    self.grids.append(grid_rband)
                else:
                    # 一度生成したインスタンスは消去せず再利用する
                    grid_rband = self.grids[grids_counter - 1]

                # 座標をセット
                for v in grid_vertexes:
                    grid_rband.addPoint(v, True)
                # 縁を赤色に、背景色を透明に
                grid_rband.setColor(QColor(255, 0, 0))
                grid_rband.setFillColor(QColor(0, 0, 0, 0))
                grid_rband.show()

    def calc_xy_grids_count(self, currentpoint: QPoint):
        """[summary]
        startpointを基準に、xy方向にグリッドをいくつ表示すべきか計算
        Args:
            currentpoint (QPoint): ドラッグ中のキャンバス座標

        Returns:
            [int, int]: [x方向のグリッドの数, y方向のグリッドの数]
        """

        local_startpoint = self.startpoint
        local_currentpoint = currentpoint
        point_extent = currentpoint - local_startpoint
        if self.centerMode:
            # if type(self.startpoint.x) == int:
            #    local_startpoint = QPoint(self.startpoint.x, self.startpoint.y)
            local_currentpoint = local_startpoint + point_extent
            local_startpoint = local_startpoint - point_extent

        start_coordinates = self.canvas.getCoordinateTransform(
        ).toMapCoordinates(local_startpoint.x(), local_startpoint.y())
        current_coordinates = self.canvas.getCoordinateTransform(
        ).toMapCoordinates(local_currentpoint.x(), local_currentpoint.y())
        x_sign = 1 if (current_coordinates.x() - start_coordinates.x()) > 0 else -1
        y_sign = 1 if (current_coordinates.y() - start_coordinates.y()) > 0 else -1
        area_width = current_coordinates.x() - start_coordinates.x()
        area_height = current_coordinates.y() - start_coordinates.y()
        x_grids_count = int(area_width / self.tilesize[0])
        y_grids_count = int(area_height / self.tilesize[1])
        return x_grids_count, y_grids_count, x_sign, y_sign

    def sort_as_lefttop_rightbottom(self, point1, point2):
        """
        1組のQPointの入力を左上xy・右下xyに変換する

        Args:
            point1 (QPoint, QgsPointXY): 座標1
            point2 (QPoint, QgsPointXY): 座標2
            いずれもx(), y()が実装されていること

        Returns:
            [int, int], [int, int]: 左上xy、右下xyの座標
        """
        x_point1 = point1.x()
        y_point1 = point1.y()
        x_point2 = point2.x()
        y_point2 = point2.y()

        left = min(x_point1, x_point2)
        right = max(x_point1, x_point2)
        bottom = min(y_point1, y_point2)
        top = max(y_point1, y_point2)

        return [left, top], [right, bottom]

    def calc_quadrant(self, start_coordinates: QgsPointXY, end_coordinates: QgsPointXY) -> int:
        """
        ドラッグ開始点から見た終了点が属する象限を算出する

        Args:
            start_coordinates (QgsPointXY): ドラッグ開始点
            end_coordinates (QgsPointXY): ドラッグ終了点

        Returns:
            int: 象限
        """

        is_rightside = start_coordinates.x() < end_coordinates.x()
        is_upperside = start_coordinates.y() < end_coordinates.y()

        if is_rightside:
            if is_upperside:
                return 1
            else:
                return 4
        else:
            if is_upperside:
                return 2
            else:
                return 3
